from ...general_operations.geometry_intersections.rectangle_with_rectangle import rectangle_rectangle_intersection
from ...problem_operations.conflicts.blocked_magnet import conflicted_magnet

# check for conflicts between two rectangular magnets by checking each of the pickup direction of one magnet with the other
def check_conflict_rectangle_rectangle_magnets(magnet_pair):

    all_blocked_pickup_areas = []

    # for checking the magnets against each other
    for magnet_1, magnet_2 in zip([0, 1], [1, 0]):

        # for checking through each of the selected magnet's pickup areas
        for pickup_area in magnet_pair[magnet_1].pickup_areas:

            # creating points if there is intersection between the respective pickup area of each magnet with the other
            points = rectangle_rectangle_intersection(magnet_pair[magnet_2], pickup_area)

            # if points of intersection exist, the respective pickup area of the respective magnet is blocked
            if points:

                # adding the identified blocked pickup area
                all_blocked_pickup_areas.append(conflicted_magnet(magnet_pair[magnet_2], magnet_pair[magnet_1], pickup_area))

                ## PLOTS TO VISUALIZE THE INTERSECTIONS FOR CLOSE INSPECTION
                # plt.figure()
                # magnet_pair[magnet_1].draw_rectangle('c')
                # plt.text(magnet_pair[magnet_1].center[0], magnet_pair[magnet_1].center[1], str(int(magnet_pair[magnet_1].index)), fontsize=6)
                #
                # magnet_pair[magnet_2].draw_rectangle('r')
                # plt.text(magnet_pair[magnet_2].center[0], magnet_pair[magnet_2].center[1], str(int(magnet_pair[magnet_2].index)), fontsize=6)
                #
                # pickup_area.draw_rectangle('r')

                # for point in points:
                #    plt.plot(point[0],point[1],'or')

    return all_blocked_pickup_areas






from ...hector.constants import rectangle_magnet_length, robot_arm_width,robot_arm_length,circular_magnet_radius
from ...general_operations.geometry_shapes.rectangle import rectangle

class rectangular_magnet_pickup_area(rectangle):

    # intializing pickup area of rectangular magnet with the respective parameters
    def __init__(self,center,orientation):

        #length = 0.5 * (rectangle_magnet_length + 3* robot_arm_width) # Tiphaine's calculation
        if (robot_arm_width < ((rectangle_magnet_length - robot_arm_width) / 2)):
            length = (robot_arm_width + rectangle_magnet_length) / 2
        elif (robot_arm_width >= ((rectangle_magnet_length - robot_arm_width) / 2)):
            length = 2 * robot_arm_width

        width  = robot_arm_length
        super().__init__(center,length,width,orientation)


class circular_magnet_pickup_area(rectangle):

    # intializing pickup area of rectangular magnet with the respective parameters
    def __init__(self,center,orientation):
        length = (3.0/2) * robot_arm_width + circular_magnet_radius
        width  = robot_arm_length
        super().__init__(center,length,width,orientation)


## Pickup Areas of circular and rectangular magnets created as classes

class inward(rectangular_magnet_pickup_area):
    kind = 'inward (toward circular magnet) pickup area'
    code = 'I'
    pass

class outward(rectangular_magnet_pickup_area):
    kind = 'outward (toward circular magnet) pickup area'
    code = 'O'
    pass

class tangential_right(circular_magnet_pickup_area):
    kind = 'tangential from right pickup area'
    code = 'TR'
    pass

class tangential_left(circular_magnet_pickup_area):
    kind = 'tangential from left pickup area'
    code = 'TL'
    pass

class radial_inward(circular_magnet_pickup_area):
    kind = 'inward radial pickup area'
    code ='RI'
    pass

class radial_outward(circular_magnet_pickup_area):
    kind = 'outward radial pickup area'
    code = 'RO'
    pass






from ...general_operations.geometry_intersections.segment_with_segment import segments_intersection
from ...general_operations.geometry_shapes.segment import segment

# calculating intersection points between two rectangular magnets
def rectangle_rectangle_intersection(rectangle1, rectangle2):

    # calculate four corners of the rectangular magnet
    rectangle1_corners = rectangle1.calculate_4corners()
    rectangle2_corners = rectangle2.calculate_4corners()

    all_intersection_points = []

    for k,l in zip([0, 1, 2, 3],[1, 2, 3, 0]):
        for m,n in zip([0, 1, 2, 3],[1, 2, 3, 0]):

            # creating rectangle magnet sides using rectangular magnet corners
            rectangle1_side = segment(rectangle1_corners[k], rectangle1_corners[l])
            rectangle2_side = segment(rectangle2_corners[m], rectangle2_corners[n])

            # solving for intersection points between two rectangular segments
            point = segments_intersection(rectangle1_side,rectangle2_side)

            # adding all intersecting points is list
            if point:
                all_intersection_points.append(point)

    return all_intersection_points



from astropy.io import fits 
from astropy.table import Table

def load_FITS_table_in_pandas(fname):

    hdu=fits.open(fname)
    t=Table(hdu[1].data)
    return t.to_pandas()

def save_dataframe_as_FITS_table(df, fname, overwrite=True):

    t=Table.from_pandas(df)
    return t.write(fname, overwrite=overwrite)


from ...problem_operations.conflicts.blocked_magnet import conflicted_magnet
from ...general_operations.geometry_intersections.circle_with_rectangle import circle_rectangle_intersection
from ...general_operations.geometry_intersections.rectangle_with_rectangle import rectangle_rectangle_intersection
from ...hector.magnets.circular import is_circular_magnet
from ...hector.magnets.rectangular import is_rectangular_magnet

# check for the circular magnet blocking the rectangular magnet
def circle_blocking_rectangle(circle,rectangle):

    all_blocked_pickup_areas = []

    # for checking through each of rectangular magnet's pickup areas
    for pickup_area in rectangle.pickup_areas:

        # creating points if there is intersection between the respective pickup area of rectangular magnet with circular one
        points = circle_rectangle_intersection(circle, pickup_area)

        # if points of intersection exist, the respective pickup area of the respective magnet is blocked
        if points:

            # adding the identified blocked pickup area
            all_blocked_pickup_areas.append(conflicted_magnet(circle, rectangle, pickup_area))

            ## PLOTS TO VISUALIZE THE INTERSECTIONS FOR CLOSE INSPECTION
            # plt.figure()
            # circle.draw_circle('r')
            # plt.text(circle.center[0], circle.center[1],str(int(circle.index)), fontsize=6)
            #
            # rectangle.draw_rectangle('c')
            # plt.text(rectangle.center[0], rectangle.center[1],str(int(rectangle.index)), fontsize=6)
            #
            # pickup_area.draw_rectangle('r')

            # for point in points:
            #    plt.plot(point[0],point[1],'or')

    return all_blocked_pickup_areas

# check for the rectangular magnet blocking the circular magnet
def rectangle_blocking_circle(circle,rectangle):

    all_blocked_pickup_areas = []

    # for checking through each of circular magnet's pickup areas
    for pickup_area in circle.pickup_areas:

        # creating points if there is intersection between the respective pickup area of circular magnet with rectangular one
        points = rectangle_rectangle_intersection(rectangle, pickup_area)

        # if points of intersection exist, the respective pickup area of the respective magnet is blocked
        if points:

            # adding the identified blocked pickup area
            all_blocked_pickup_areas.append(conflicted_magnet(rectangle, circle, pickup_area))

            ## PLOTS TO VISUALIZE THE INTERSECTIONS FOR CLOSE INSPECTION
            # plt.figure()
            # circle.draw_circle('r')
            # plt.text(circle.center[0], circle.center[1],str(int(circle.index)), fontsize=6)
            #
            # rectangle.draw_rectangle('c')
            # plt.text(rectangle.center[0], rectangle.center[1],str(int(rectangle.index)), fontsize=6)
            #
            # pickup_area.draw_rectangle('r')

            # for point in points:
            #   plt.plot(point[0],point[1],'or')

    return all_blocked_pickup_areas

# check for conflicts between circular and rectangular magnets by checking each of the pickup
# direction of one magnet with the other
def check_conflict_circle_rectangle_magnets(magnet_pair):

    all_blocked_pickup_areas = []

    # for checking the magnets against each other
    for magnet_1, magnet_2 in zip([0, 1], [1, 0]):

        # for checking through each of the circular magnet's pickup areas
        if is_circular_magnet(magnet_pair[magnet_1]):

            blocked_pickup_area = circle_blocking_rectangle(magnet_pair[magnet_1], magnet_pair[magnet_2])

            # adding the identified blocked pickup area, if any
            if blocked_pickup_area:

                all_blocked_pickup_areas.extend(blocked_pickup_area)

        # for checking through each of the rectangular magnet's pickup areas
        elif is_rectangular_magnet(magnet_pair[magnet_1]):

            blocked_pickup_area = rectangle_blocking_circle(magnet_pair[magnet_2], magnet_pair[magnet_1])

            # adding the identified blocked pickup area, if any
            if blocked_pickup_area:

                all_blocked_pickup_areas.extend(blocked_pickup_area)

    return all_blocked_pickup_areas



from ...hector.magnets.circular import is_circular_magnet
from ...hector.magnets.rectangular import is_rectangular_magnet

# check for the pair of magnets if both are circular
def circle_and_circle_magnets(pair):

    return is_circular_magnet(pair[0]) and is_circular_magnet(pair[1])

# check for the pair of magnets if both are rectangular
def circle_rectangle_magnets(pair):

    if is_circular_magnet(pair[0]) and is_rectangular_magnet(pair[1]):
        return True

    elif is_rectangular_magnet(pair[0]) and is_circular_magnet(pair[1]):
        return True

    else:
        return False

# check for the pair of magnets if one is circular and the other rectangular
def rectangle_rectangle_magnets(pair):

    return is_rectangular_magnet(pair[0]) and is_rectangular_magnet(pair[1])


#!/usr/bin/env python3
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import csv
import time
start_time = time.time()

from hop.hexabundle_allocation.problem_operations import extract_data, file_arranging, hexabundle, offsets, plots, \
                                                        position_ordering, robot_parameters, conflicts
from hop.hexabundle_allocation.hector.plate import HECTOR_plate

clusterNum = 2
tileNum = 0
limit = 86
allBatchOfTiles = ['G9','G12','G15']

Location = 'D:/Hector/Hector_project_files/Hector/magnet_position'
galaxyIDrecord = {}

def allocate_hexas(clusterNum, tileNum, galaxyIDrecord, plot):

    # check for galaxyIDrecord python dictionary
    if 'galaxyIDrecord' in globals() or 'galaxyIDrecord' in locals():
        print("galaxyIDrecord exists.")
    else:
        print("galaxyIDrecord being created.")
        galaxyIDrecord = {}

    # fileNameGuides = ('GAMA_'+batch+'/Configuration/HECTORConfig_Guides_GAMA_'+batch+'_tile_%03d.txt' % (tileNum))
    fileNameGuides = (Location+'/For_Ayoan_14p5_exclusion_Clusters/Cluster_%d/Configuration/HECTORConfig_Guides_Cluster_%d_tile_%03d.txt' % ( clusterNum, clusterNum, tileNum))

    # proxy file to arrange guides in required format to merge with hexa probes
    proxyGuideFile = Location+'/galaxy_fields/newfile.txt'

    # Adding ID column and getting rid of the header line of Guides cluster to add to the hexa cluster
    file_arranging.arrange_guidesFile(fileNameGuides, proxyGuideFile)

    # fileNameHexa = ('GAMA_'+batch+'/Configuration/HECTORConfig_Hexa_GAMA_'+batch+'_tile_%03d.txt' % (tileNum))
    fileNameHexa = (Location+'/For_Ayoan_14p5_exclusion_Clusters/Cluster_%d/Configuration/HECTORConfig_Hexa_Cluster_%d_tile_%03d.txt' % (clusterNum, clusterNum, tileNum))

    plate_file = (Location+'/For_Ayoan_14p5_exclusion_Clusters/Cluster_%d/Output/Hexa_and_Guides_Cluster_%d_tile_%03d.txt' % (clusterNum, clusterNum, tileNum))
    # plate_file = get_file('GAMA_'+batch+'/Output/Hexa_and_Guides_GAMA_'+batch+'_tile_%03d.txt' % (tileNum))

    #### Offset function: thermal coefficient based movement of magnet pair as a whole
    plate_file,magnetPair_offset = offsets.magnetPair_radialPositionOffset(plate_file)

    # Adding guides cluster txt file to hexa cluster txt file
    file_arranging.merge_hexaAndGuides(fileNameHexa, proxyGuideFile, plate_file)

    # extracting all the magnets and making a list of them from the plate_file
    all_magnets = extract_data.create_list_of_all_magnets_from_file(extract_data.get_file(plate_file),magnetPair_offset)

    #### Offset functions- still a work in progress- need to determine input source and add column to output file
    all_magnets = offsets.hexaPositionOffset(all_magnets)

    # create magnet pickup areas for all the magnets
    plots.create_magnet_pickup_areas(all_magnets)

    if plot:
        # ************** # creating plots and drawing pickup areas
        plt.clf()
        plt.close()
        HECTOR_plate().draw_circle('r')
        plots.draw_magnet_pickup_areas(all_magnets, '--c')
        # **************


    # test for collision and detect magnets which have all pickup directions blocked
    conflicted_magnets = conflicts.functions.find_all_blocked_magnets(all_magnets)

    # create a list of the fully blocked magnets
    fully_blocked_magnets = conflicts.functions.create_list_of_fully_blocked_magnets(conflicted_magnets)

    conflicts.functions.blocking_magnets_for_fully_blocked_magnets(conflicted_magnets)

    # print the fully blocked magnets out in terminal and record in conflicts file
    conflictsRecord = Location+'/galaxy_fields/Conflicts_Index.txt'
    conflicts.blocked_magnet.print_fully_blocked_magnets(fully_blocked_magnets,conflictsRecord, fileNameHexa)

    conflictFile = Location+'/galaxy_fields/unresolvable_conflicts.txt'
    flagsFile = Location+'/galaxy_fields/Flags.txt'
    #***  Choose former method OR median method OR larger bundle prioritized method for hexabundle allocation  ***
    positioning_array,galaxyIDrecord = position_ordering.create_position_ordering_array(all_magnets, fully_blocked_magnets, \
                                  conflicted_magnets, galaxyIDrecord, clusterNum, tileNum, conflictFile, flagsFile)

    if plot:
        # draw all the magnets in the plots created earlier
        figureFile = (Location+'/figures/Cluster_%d/savedPlot_cluster_%d_tile_%03d.pdf' % (clusterNum,clusterNum,tileNum))
        plots.draw_all_magnets(all_magnets,clusterNum,tileNum,figureFile)  #***********

    # checking positioning_array prints out all desired parameters
    print(positioning_array)

    # insert column heading and print only rectangular magnet rows in the csv file
    newrow = ['Magnet', 'Label', 'Center_x', 'Center_y', 'rot_holdingPosition', 'rot_platePlacing', 'order', 'Pickup_option', 'ID','Index', 'Hexabundle']
    newrow_circular = ['Magnet', 'Label', 'Center_x', 'Center_y', 'holding_position_ang', 'plate_placement_ang', 'order', 'Pickup_option', 'ID', 'Index', 'Hexabundle']

    # final two output files
    outputFile = (Location+'/For_Ayoan_14p5_exclusion_Clusters/Cluster_%d/Output_with_Positioning_array/Hexa_and_Guides_with_PositioningArray_Cluster_%d_tile_%03d.txt' \
                % (clusterNum, clusterNum, tileNum))
                # ('GAMA_'+batch+'/Output_with_Positioning_array/Hexa_and_Guides_with_PositioningArray_GAMA_'+batch+'_tile_%03d.txt' % (tileNum)
    robotFile = (Location+'/For_Ayoan_14p5_exclusion_Clusters/Cluster_%d/Output_for_Robot/Robot_Cluster_%d_tile_%03d.txt' \
                 % (clusterNum, clusterNum, tileNum))
                #('GAMA_'+batch+'/Output_for_Robot/Robot_GAMA_'+batch+'_tile_%03d.txt' % (tileNum)

    # creating robotFile array and storing it in robot file
    positioning_array, robotFilearray = file_arranging.create_robotFileArray(positioning_array,robotFile,newrow)

    # adjusting the positioning array to merge only selected parameters to the output file
    positioning_array, positioning_array_circular = file_arranging.positioningArray_adjust_and_mergetoFile(positioning_array, plate_file, outputFile, newrow,newrow_circular)

    # produce final files with consistent layout and no extra commas
    file_arranging.finalFiles(outputFile, robotFile)

    print(galaxyIDrecord)

    # just to check each tile's whole operation time
    print("\t \t -----   %s seconds   -----" % (time.time() - start_time))

    # Comment out all ***** marked plot functions above(lines 81-84,105s)
    # to run whole batch of tiles fast without plots


# for batch in allBatchOfTiles:
#     if batch == 'G9':
#         limit = 104
#     elif batch == 'G12':
#         limit = 86
#     elif batch == 'G15':
#         limit = 95


for clusterNum in range(2,3):
    if (clusterNum == 1):
        limit = 15
    elif (clusterNum == 2):
        limit = 29
    elif (clusterNum == 3):
        limit = 24
    elif (clusterNum == 4):
        limit = 10
    elif (clusterNum == 5):
        limit = 26
    elif (clusterNum == 6):
        limit = 16
    for tileNum in range(21,22):
        allocate_hexas(clusterNum,tileNum,galaxyIDrecord, plot=True)


#!/usr/bin/env python3
import sys
sys.path.append('/home/ayoan/Hector/magnet_position/venv/bin/python')

import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import csv
import time
start_time = time.time()

from problem_operations.conflicts.functions import find_all_blocked_magnets,\
                                    create_list_of_fully_blocked_magnets
from problem_operations.conflicts.blocked_magnet import print_fully_blocked_magnets
from problem_operations.extract_data import create_list_of_all_magnets_from_file,get_file
from problem_operations.plots import create_magnet_pickup_areas, draw_magnet_pickup_areas, draw_all_magnets
from problem_operations.position_ordering import create_position_ordering_array
from problem_operations.file_arranging import arrange_guidesFile, merge_hexaAndGuides, create_robotFileArray, \
                                    positioningArray_adjust_and_mergetoFile, finalFiles
from hector.plate import HECTOR_plate
from problem_operations.offsets import hexaPositionOffset


clusterNum = 2
tileNum = 0
limit = 86
allBatchOfTiles = ['G9','G12','G15']

galaxyIDrecord = {}

# for batch in allBatchOfTiles:
#     if batch == 'G9':
#         limit = 104
#     elif batch == 'G12':
#         limit = 86
#     elif batch == 'G15':
#         limit = 95
for clusterNum in range(2,3):
    if (clusterNum == 1):
        limit = 15
    elif (clusterNum == 2):
        limit = 29
    elif (clusterNum == 3):
        limit = 24
    elif (clusterNum == 4):
        limit = 10
    elif (clusterNum == 5):
        limit = 26
    elif (clusterNum == 6):
        limit = 16
    for tileNum in range(7,10):

        # fileNameGuides = ('GAMA_'+batch+'/Configuration/HECTORConfig_Guides_GAMA_'+batch+'_tile_%03d.txt' % (tileNum))
        fileNameGuides = ('For_Ayoan_14p5_exclusion_Clusters/Cluster_%d/Configuration/HECTORConfig_Guides_Cluster_%d_tile_%03d.txt' % (clusterNum, clusterNum, tileNum))

        # proxy file to arrange guides in required format to merge with hexa probes
        proxyGuideFile = 'galaxy_fields/newfile.txt'

        # Adding ID column and getting rid of the header line of Guides cluster to add to the hexa cluster
        arrange_guidesFile(fileNameGuides, proxyGuideFile)

        # fileNameHexa = ('GAMA_'+batch+'/Configuration/HECTORConfig_Hexa_GAMA_'+batch+'_tile_%03d.txt' % (tileNum))
        fileNameHexa = ('For_Ayoan_14p5_exclusion_Clusters/Cluster_%d/Configuration/HECTORConfig_Hexa_Cluster_%d_tile_%03d.txt' % (clusterNum, clusterNum, tileNum))

        plate_file = ('For_Ayoan_14p5_exclusion_Clusters/Cluster_%d/Output/Hexa_and_Guides_Cluster_%d_tile_%03d.txt' % (clusterNum, clusterNum, tileNum))
        # plate_file = get_file('GAMA_'+batch+'/Output/Hexa_and_Guides_GAMA_'+batch+'_tile_%03d.txt' % (tileNum))

        # Adding guides cluster txt file to hexa cluster txt file
        merge_hexaAndGuides(fileNameHexa, proxyGuideFile, plate_file, clusterNum, tileNum)

        # extracting all the magnets and making a list of them from the plate_file
        all_magnets = create_list_of_all_magnets_from_file(get_file(plate_file))

        #### Offset functions- still a work in progress- need to determine input source and add column to output file
        all_magnets = hexaPositionOffset(all_magnets)

        # create magnet pickup areas for all the magnets
        create_magnet_pickup_areas(all_magnets)

        #************** # creating plots and drawing pickup areas
        plt.clf()
        plt.close()
        HECTOR_plate().draw_circle('r')
        draw_magnet_pickup_areas(all_magnets, '--c')
        #**************


        # test for collision and detect magnets which have all pickup directions blocked
        conflicted_magnets = find_all_blocked_magnets(all_magnets)

        # create a list of the fully blocked magnets
        fully_blocked_magnets = create_list_of_fully_blocked_magnets(conflicted_magnets)

        # print the fully blocked magnets out in terminal and record in conflicts file
        conflictsRecord = 'galaxy_fields/Conflicts_Index.txt'
        print_fully_blocked_magnets(fully_blocked_magnets,conflictsRecord, fileNameHexa)

        conflictFile = 'galaxy_fields/unresolvable_conflicts.txt'
        flagsFile = 'galaxy_fields/Flags.txt'
        #***  Choose former method OR median method OR larger bundle prioritized method for hexabundle allocation  ***
        positioning_array,galaxyIDrecord = create_position_ordering_array(all_magnets, fully_blocked_magnets, \
                                      conflicted_magnets, galaxyIDrecord, clusterNum, tileNum, conflictFile, flagsFile)

        # draw all the magnets in the plots created earlier
        figureFile = ('figures/Cluster_%d/savedPlot_cluster_%d_tile_%03d.pdf' % (clusterNum,clusterNum,tileNum))
        draw_all_magnets(all_magnets,clusterNum,tileNum,figureFile)  #***********

        # checking positioning_array prints out all desired parameters
        print(positioning_array)

        # insert column heading and print only rectangular magnet rows in the csv file
        newrow = ['Magnet', 'Label', 'Center_x', 'Center_y', 'rot_holdingPosition', 'rot_platePlacing', 'order', 'Pickup_option', 'ID','Index', 'Hexabundle']
        newrow_circular = ['Magnet', 'Label', 'Center_x', 'Center_y', 'holding_position_ang', 'plate_placement_ang', 'order', 'Pickup_option', 'ID', 'Index', 'Hexabundle']

        # final two output files
        outputFile = ('For_Ayoan_14p5_exclusion_Clusters/Cluster_%d/Output_with_Positioning_array/Hexa_and_Guides_with_PositioningArray_Cluster_%d_tile_%03d.txt' \
                    % (clusterNum, clusterNum, tileNum))
                    # ('GAMA_'+batch+'/Output_with_Positioning_array/Hexa_and_Guides_with_PositioningArray_GAMA_'+batch+'_tile_%03d.txt' % (tileNum)
        robotFile = ('For_Ayoan_14p5_exclusion_Clusters/Cluster_%d/Output_for_Robot/Robot_Cluster_%d_tile_%03d.txt' \
                     % (clusterNum, clusterNum, tileNum))
                    #('GAMA_'+batch+'/Output_for_Robot/Robot_GAMA_'+batch+'_tile_%03d.txt' % (tileNum)

        # creating robotFile array and storing it in robot file
        positioning_array, robotFilearray = create_robotFileArray(positioning_array,robotFile,newrow)

        # adjusting the positioning array to merge only selected parameters to the output file
        positioning_array, positioning_array_circular = positioningArray_adjust_and_mergetoFile(positioning_array, \
                                                                        plate_file, outputFile, newrow,newrow_circular)

        # produce final files with consistent layout and no extra commas
        finalFiles(outputFile, robotFile)

        # just to check each tile's whole operation time
        print("\t \t -----   %s seconds   -----" % (time.time() - start_time))


        # Comment out all ***** marked plot functions above(lines 81-84,105s)
        # to run whole batch of tiles fast without plots



from ...hector.magnets.magnet_pair import circle_and_circle_magnets, circle_rectangle_magnets, rectangle_rectangle_magnets
from .circular_magnet_with_circular_magnet import check_conflict_circle_circle_magnets
from .circular_magnet_with_rectangular_magnet import check_conflict_circle_rectangle_magnets
from .rectangular_magnet_and_rectangular_magnet import check_conflict_rectangle_rectangle_magnets
from .blocked_magnet import circular_magnet_is_fully_blocked,rectangular_magnet_is_fully_blocked
from ...hector.constants import rectangle_magnet_length, robot_arm_width
import numpy as np
from ...hector.magnets.rectangular import rectangular_magnet, is_rectangular_magnet
from ...hector.magnets.circular import circular_magnet, is_circular_magnet

## The function names with no comments have self-defining names

def minimum_magnet_proximity():
    return rectangle_magnet_length + robot_arm_width

def calculate_magnet_to_magnet_distance(magnet1, magnet2):
    return np.sqrt((magnet2.center[0] - magnet1.center[0]) ** 2 + (magnet2.center[1] - magnet1.center[1]) ** 2)

def create_list_magnets_in_close_proximity(list_magnets):

    magnets_in_close_proximity = []

    # checking each magnet with all the other ones iteratively through the magnets list
    for i in range(len(list_magnets)):
        magnet_i = list_magnets[i]

        for j in range(i+1, len(list_magnets)):
            magnet_j = list_magnets[j]

            # adding to list if distance between selected magnets are less than the specified minimum proximity
            if calculate_magnet_to_magnet_distance(magnet_i, magnet_j) < minimum_magnet_proximity():
                magnets_in_close_proximity.append([magnet_i, magnet_j])

    return np.array(magnets_in_close_proximity)

def highlight_closed_magnets_on_plot(closed_magnets, colour):

    # iterating through magnet pairs identified as closed magnets
    for magnet_pair in closed_magnets:

        # iterating through each magnet of the selected pair
        for magnet in magnet_pair:

            # highlighting the respective rectangular magnet with provided input colour
            if isinstance(magnet, rectangular_magnet):
                magnet.draw_rectangle(colour)

            # highlighting the respective circular magnet with provided input colour
            elif isinstance(magnet, circular_magnet):
                magnet.draw_circle(colour)

def find_conflicts_between_magnets(pair):

    all_blocked_pickup_areas = []

    # checking if the magnet pair consists of both circular ones
    if circle_and_circle_magnets(pair):

        # check for conflict between circular magnets
        magnet_conflict = check_conflict_circle_circle_magnets(pair)

        # record the identified conflicts as blocked pickup areas
        if magnet_conflict:
            all_blocked_pickup_areas.extend(magnet_conflict)

    # checking if the magnet pair consists of one circular and one rectangular
    elif circle_rectangle_magnets(pair):

        # check for conflict between circular and rectangular magnets
        magnet_conflict = check_conflict_circle_rectangle_magnets(pair)

        # record the identified conflicts as blocked pickup areas
        if magnet_conflict:
            all_blocked_pickup_areas.extend(magnet_conflict)

    # checking if the magnet pair consists of both rectangular ones
    elif rectangle_rectangle_magnets(pair):

        # check for conflict between rectangular magnets
        magnet_conflict = check_conflict_rectangle_rectangle_magnets(pair)

        # record the identified conflicts as blocked pickup areas
        if magnet_conflict:
            all_blocked_pickup_areas.extend(magnet_conflict)

    return all_blocked_pickup_areas

def find_all_blocked_magnets(all_magnets):

    # create magnet pairs from magnets list which have specified close proximity
    magnet_pairs = create_list_magnets_in_close_proximity(all_magnets)

    blocked_areas = []

    for pair in magnet_pairs:

        # check for conflict between the magnet pairs in specified close proximity
        magnet_conflict = find_conflicts_between_magnets(pair)

        # check if magnet conflicts exist
        if len(magnet_conflict) > 0:

            # add the magnet conflicted area to the blocked areas
            blocked_areas.extend(magnet_conflict)

            # remove the blocked pickup areas from list of available pickup by iterating through conflicts
            for conflict in magnet_conflict:
                conflict.remove_blocked_pickup_areas_from_list_available_pickup()

    return blocked_areas

def is_magnet_fully_blocked(magnet,blocked_pickup_areas):

    # check for circular magnet being fully blocked
    if is_circular_magnet(magnet):
        magnet_is_fully_blocked = circular_magnet_is_fully_blocked(blocked_pickup_areas)

    # check for rectangular magnet being fully blocked
    elif is_rectangular_magnet(magnet):
        magnet_is_fully_blocked = rectangular_magnet_is_fully_blocked(blocked_pickup_areas)

    return magnet_is_fully_blocked

def create_list_of_blocked_pickup_areas(magnet, conflicted_magnet_list):

    blocked_pickup_areas = []

    for conflict in conflicted_magnet_list:

        # identify and add the blocked pickup areas from the conflicts by checking against the blocked magnets
        if magnet == conflict.blocked_magnet:
            blocked_pickup_areas.append(conflict.blocked_pickup_area)

    return blocked_pickup_areas

def remove_multiple_occurrences_in_list(object_list):
    return list(set(object_list))

def create_list_of_fully_blocked_magnets(list_of_blocked_magnets):

    fully_blocked_magnets = []

    for conflict in list_of_blocked_magnets:

        # creating list of blocked pickup areas by checking each blocked magnet against blocked magnets list
        blocked_pickup_areas = create_list_of_blocked_pickup_areas(conflict.blocked_magnet, list_of_blocked_magnets)

        # identify and add fully blocked magnets by checking them against blocked pickup areas
        if is_magnet_fully_blocked(conflict.blocked_magnet,blocked_pickup_areas):
            fully_blocked_magnets.append(conflict.blocked_magnet)

    return remove_multiple_occurrences_in_list(fully_blocked_magnets)

# identifying the magnets which are responsible for blocking each of the fully blocked magnets
def blocking_magnets_for_fully_blocked_magnets(list_of_blocked_magnets):

    fully_blocked_magnets_dictionary = {}

    for conflict in list_of_blocked_magnets:

        blocked_pickup_areas = create_list_of_blocked_pickup_areas(conflict.blocked_magnet, list_of_blocked_magnets)

        if is_magnet_fully_blocked(conflict.blocked_magnet, blocked_pickup_areas):

            # magnet conflicts stated out with each fully blocked magnet and its list of blocking magnets
            conflict_magnet = 'For ' + conflict.blocked_magnet.__class__.__name__ + ' ' + str(int(conflict.blocked_magnet.index)) + ': ' + \
                            conflict.blocking_magnet.__class__.__name__ + ' ' + str(int(conflict.blocking_magnet.index)) + '\n'

            ## TEST PRINT
            print(conflict_magnet)

            # blocked magnet stored as key and the blocking magnets as values in the fully blocked magnets dictionary
            key = conflict.blocked_magnet.__class__.__name__ + ' ' + str(int(conflict.blocked_magnet.index))
            value = conflict.blocking_magnet.__class__.__name__ + ' ' + str(int(conflict.blocking_magnet.index))

            # store the values by creating the keys if necessary and adding multiple values to same key if required
            fully_blocked_magnets_dictionary.setdefault(key, [])
            if value not in fully_blocked_magnets_dictionary[key]:
                fully_blocked_magnets_dictionary[key].append(value)

    ## TEST PRINT
    print(fully_blocked_magnets_dictionary)

    return fully_blocked_magnets_dictionary

def create_list_of_blocking_magnets(list_of_conflicts, blocked_magnet):

    blocking_magnets = []

    for conflict in list_of_conflicts:

        # check for the blocked magnet in list of conflicts and add the respective blocking magnets
        if conflict.blocked_magnet == blocked_magnet:
            blocking_magnets.append(conflict.blocking_magnet)

    return remove_multiple_occurrences_in_list(blocking_magnets)

def blocking_magnet_is_fully_blocked(blocking_magnet, list_of_fully_blocked_magnets):
    return blocking_magnet in list_of_fully_blocked_magnets

def all_blocking_magnets_are_fully_blocked(list_of_blocking_magnets,list_of_fully_blocked_magnets):

    all_blocking_magnets_are_fully_blocked = True

    # check against each blocking magnet whether all the blocking magnets are fully blocked
    for blocking_magnet in list_of_blocking_magnets:
        all_blocking_magnets_are_fully_blocked *= blocking_magnet_is_fully_blocked(blocking_magnet, list_of_fully_blocked_magnets)

    return all_blocking_magnets_are_fully_blocked



import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
from ..hexabundle_allocation.hector import constants as hector_constants

def plot_distortion_correction_before_after(DC_tile_filename, title_text=None):

    """
    Take a tile which has had the distortion correction applied. Plot a before/after image to see the effects
    """

    font_size = 12

    # Get the RA and DEC of the tile
    with open(DC_tile_filename, 'r') as f:
        lines = f.readlines()
    RA, DEC = lines[1].split()[1:]
    RA = float(RA)
    DEC = float(DEC)

    df = pd.read_csv(DC_tile_filename, skiprows=8)

    # Select only galaxies in the tile
    mask = df.type == 1

    # Rename things for brevity
    x_noDC = df.loc[mask, 'MagnetX_noDC']/1000
    y_noDC = df.loc[mask, 'MagnetY_noDC']/1000

    x_DC = df.loc[mask, 'MagnetX']/1000
    y_DC = df.loc[mask, 'MagnetY']/1000

    # Find the offsets for plotting
    dxs = np.row_stack([x_noDC, x_DC])
    dys = np.row_stack([y_noDC, y_DC])

    # Do the plotting
    fig, ax = plt.subplots(figsize=(8, 8))

    ax.plot(dxs, dys, c='k', zorder=1)
    ax.scatter(x_noDC, y_noDC, label='Before correction', c='b', s=40, zorder=10)
    ax.scatter(x_DC, y_DC, label='After correction', c='r', s=40, zorder=10)

    Hector_FOV = plt.Circle((0, 0), radius=hector_constants.HECTOR_plate_radius, linestyle='dashed', color='k', fill=False)

    ax.add_patch(Hector_FOV)
    ax.set_xlim(-1.2 * hector_constants.HECTOR_plate_radius, 1.2 * hector_constants.HECTOR_plate_radius)
    ax.set_ylim(-1.2 * hector_constants.HECTOR_plate_radius, 1.2 * hector_constants.HECTOR_plate_radius)

    ax.scatter(0, 0, marker='+', c='k')
    ax.legend(fontsize=font_size, loc='lower left')
    #ax.set_box_aspect(1)

    ax.set_xlabel("X (mm)")
    ax.set_ylabel("Y (mm)")

    ax.set_title(title_text, fontsize=font_size)

    box_props = dict(boxstyle='round', facecolor='0.9', alpha=0.5)
    ax.text(0.5, 0.95, s=f"Tile Centre is ({RA:.3f}, {DEC:.3f})", transform=ax.transAxes, bbox=box_props, fontsize=font_size, ha='center', va='bottom')

    return fig, ax




import numpy as np

class circle_line_intersection_equation():

    # initializing the circle line intersection equation with the respective parameters
    def __init__(self,A,B,C,line_segment):
        self.parameter_A = A
        self.parameter_B = B
        self.parameter_C = C

        self.segment = line_segment

    # calculating determinant via: b^2 - 4ac
    def calculate_determinant(self):

        self.determinant = self.parameter_B ** 2 - 4 * self.parameter_A * self.parameter_C
        return self.determinant

    # determining intersection by checking whether it's greater than or equal to zero
    def intersection(self):

        if self.calculate_determinant() >= 0:
            return True
        else:
            return False

    # solving the intersection for roots using the quadratic formula
    def solve_intersection_times(self):

        if self.intersection():
            t1 = (- self.parameter_B + np.sqrt(self.determinant)) / (2 * self.parameter_A)
            t2 = (- self.parameter_B - np.sqrt(self.determinant)) / (2 * self.parameter_A)

            return t1,t2

    # solving for intersection points by multiply roots with the vector then adding with start point
    def solve_intersecting_points(self):

        if self.intersection():
            return [[self.segment.start_point[0] + self.solve_intersection_times()[0] * self.segment.vector()[0],
                     self.segment.start_point[1] + self.solve_intersection_times()[0] * self.segment.vector()[1]],
                    [self.segment.start_point[0] + self.solve_intersection_times()[1] * self.segment.vector()[0],
                     self.segment.start_point[1] + self.solve_intersection_times()[1] * self.segment.vector()[1]]]

    # solving for all the intersection points of segment with circle
    def solve_segment_intersects_with_circle(self):

        circle_line_intersection_points = self.solve_intersecting_points()

        circle_segment_intersection_points = []

        # adding all the circle segment intersection points in list
        if circle_line_intersection_points:
            for point in circle_line_intersection_points:

                if point_is_on_segment(point, self.segment):

                    circle_segment_intersection_points.append(point)

        return circle_segment_intersection_points

# creating a circle segment equation by calculating the required parameters
def create_circle_segment_equation(circle,segment):

    # changes in x and y calculated
    dx = segment.start_point[0] - circle.center[0]
    dy = segment.start_point[1] - circle.center[1]

    parameter_A = segment.vector()[0] ** 2 + segment.vector()[1] ** 2
    parameter_B = 2 * (segment.vector()[0] * dx + segment.vector()[1] * dy)
    parameter_C = dx ** 2 + dy ** 2 - circle.radius ** 2

    return circle_line_intersection_equation(parameter_A,parameter_B,parameter_C,segment)

# checking whether point is on segment by the range of start point to end point coordinates
def point_is_on_segment(point, segment):

    if min(segment.start_point[0],segment.end_point[0]) <= point[0] <= max(segment.start_point[0],segment.end_point[0]):
        if min(segment.start_point[1],segment.end_point[1]) <= point[1] <= max(segment.start_point[1],segment.end_point[1]):
            return True
        else:
            return False
    else:
        return False




from math import pi, cos, sin

# function for rotational matrix required for rotating at the specified rotation angle
def rotational_matrix(rotation_angle):
    return [[cos(rotation_angle), - sin(rotation_angle)],
            [sin(rotation_angle),   cos(rotation_angle)]]

# function  for converting degrees to radians
def convert_degrees_to_radians(angle_degrees):
    return angle_degrees * pi/180

# function  for converting radians to degrees
def convert_radians_to_degrees(angle_radians):
    return angle_radians * 180/pi

# function to keep angle in range of 0 to 2pi
def convert_modulus_angle(angle):
    piX2 = 2 * pi

    if angle > piX2:
        angle = angle - piX2
    if angle < 0:
        angle = piX2 - abs(angle)

    return angle


import numpy as np 
import os
import glob
from pathlib import Path
import pandas as pd
import matplotlib.pyplot as plt
import subprocess
import sys

from hop.misc import misc_tools
from hop.misc import pandas_tools as P
from hop.misc import plotting_tools as plotting_tools
from hop.tiling import tiling_functions as tiling
from hop.hexabundle_allocation.problem_operations import extract_data, file_arranging, hexabundle, offsets, plots, position_ordering, robot_parameters, conflicts

from hop.hexabundle_allocation.hector.plate import HECTOR_plate

class HectorPipe:

    """
    The Hector Observations pipeline. This class has methods to:

        * Use the Survey Simulator to create an input catalogue from a larger master catalogue
        * Tile an input catalogue using the Hector tiling code
        * Apply distortion correction to each tile
        * Configure each tile
        * Run the HexabundleAllocation code
    """
    def __init__(self, config_filename):
        """
        Must be initialised with a configuration dictionary. This can be made from a configuration file using the yaml library.
        """

        # Firstly, set the directory of the 'hop' folder as a bash environment variable, so that we can access it in R
        os.environ['HECTOROBSPIPELINE_LOC'] = Path(__file__).parent.as_posix()

        self.config_filename = config_filename
        self.config = misc_tools._load_config(config_filename)

        # Lists for the tile RAs and DECs
        self.best_tile_RAs = []
        self.best_tile_Decs = []

        # Set up the output folders
        folders = misc_tools.create_output_directories(self.config['output_folder'])
        # Add these as class attributes

        ubfolders_to_be_made = ['Logs', 'Configuration', 'Tiles', 'Plots', 'DistortionCorrected', "Allocation", "Allocation/tile_outputs", "Allocation/robot_outputs"]


        self.logfile_location = folders['Logs']
        self.configuration_location = folders['Configuration']
        self.tile_location = folders['Tiles']
        self.plot_location = folders['Plots']
        self.distortion_corrected_tile_location = folders['DistortionCorrected']
        self.distortion_corrected_plot_location = folders['DistortionCorrected/Plots']
        self.allocation_files_location_base = folders['Allocation']
        self.allocation_files_location_tiles = folders['Allocation/tile_outputs']
        self.allocation_files_location_robot = folders['Allocation/robot_outputs']


        # Set up the log files
        self.logger, self.logger_R_code = misc_tools.set_up_loggers(self.config)

        # The galaxy ID dictionary for the hexabundle allocation
        self.galaxyIDrecord = {}

        # Get the location of the distortion correction executable and the R code
        self.DistortionCorrection_binary_location = Path(__file__).parent / Path("distortion_correction/HectorTranslationSoftware/Code/HectorConfigUtil")
        if not self.DistortionCorrection_binary_location.exists():
            raise NameError("The Distortion Correction binary seems to not exist")

            Path(hop.__file__).parent
        self.TdF_distortion_file_location = Path(__file__).parent / Path("distortion_correction/HectorTranslationSoftware/Code/tdFdistortion0.sds")  
        if not self.TdF_distortion_file_location.exists():
            raise FileNotFoundError("The 2dF distortion file tdFdistortion0.sds seems to not exist")
        os.environ['TDF_DISTORTION'] = self.TdF_distortion_file_location.as_posix()

        self.ConfigurationCode_location = Path(__file__).parent / Path("configuration/HECTOR_ClusterFieldsTest.R")
        if not self.ConfigurationCode_location.exists():
            raise FileNotFoundError("The Configuration Code seems to not exist")


    def load_input_catalogue(self):
        """
        Load an input target catalogue, guide star catalogue and standard star catalogue
        """
        self.df_targets = P.load_FITS_table_in_pandas(os.path.expanduser(self.config['final_catalogue_name']))
        self.df_guide_stars = P.load_FITS_table_in_pandas(os.path.expanduser(self.config['guide_star_filename']))
        self.df_standard_stars = P.load_FITS_table_in_pandas(os.path.expanduser(self.config['standard_star_filename']))

        # Add some attributes to check that we have the correct catalogues
        self.have_targets_catalogue = True
        self.have_guide_star_catalogue = True
        self.have_standard_star_catalogue = True

    def _add_columns_to_catalogues(self):

        # Add the empty columns which we'll update
        self.df_targets['ALREADY_TILED'] = False
        self.df_targets['Tile_number'] = -999

        self.df_targets['remaining_observations'] = 1
        self.df_targets = self.df_targets.loc[self.df_targets['remaining_observations'] > 0]

        self.logger.info("Adding extra columns to the target dataframe")

    def _rename_columns(self):

        """
        Rename a few columns to make sure things align. TODO: FixMe so this isn't needed!
        """

        # # Rename columns to match the ones the code expects
        # # This will be removed in future
        self.df_targets = self.df_targets.rename(columns=dict(GAL_MAG_R='r_mag', CATID='CATAID'))
        self.df_guide_stars = self.df_guide_stars.rename(columns=dict(ROWID='CoADD_ID'))
        self.df_standard_stars = self.df_standard_stars.rename(columns=dict(ROWID='CoADD_ID'))

    def find_premade_tiles(self):
        """
        Find any tiles which have already been configured in the output folders.
        """
        pre_made_and_configured_files = np.sort(glob.glob(f"{self.configuration_location}/HECTORConfig_Hexa_*.txt"))
        pre_made_tiles = np.sort(glob.glob(os.path.expanduser(f"{self.tile_location}/tile_*.fld")))

        starting_tile = 0
        for configuration_file, tile_file in zip(pre_made_and_configured_files, pre_made_tiles):
            tile_number = int(configuration_file.split('_')[-1].split('.')[0])
            tile_members = pd.read_csv(configuration_file, delim_whitespace=True)
            already_tiled_mask = self.df_targets['CATAID'].isin(tile_members.IDs)
            # Change the TILED flag for targets we've added to a tile
            self.df_targets.loc[already_tiled_mask, 'ALREADY_TILED'] = True
            # And include the tile number for each target
            self.df_targets.loc[already_tiled_mask, 'Tile_number'] = tile_number
            starting_tile +=1

            # Now get the centres of the tile
            with open(tile_file, 'r') as f:
                lines = f.readlines()[1].split()
            self.best_tile_RAs.append(float(lines[1]))
            self.best_tile_Decs.append(float(lines[2]))

        self.logger.info(f"Loaded {starting_tile} pre-made tiles from a previous run")

        return starting_tile


    def tile_field(self, configure_tiles=True, apply_distortion_correction=True, plot=True, config_timeout=None):

        """
        Tile an entire input catalogue. Optionally apply distortion correction to go from RA/DEC to locations on the plate and optionally run the configuration code to arrange the hexabundles on the plate. 
        """

        # Check that we have our three input catalogues
        if not self.have_targets_catalogue & self.have_standard_star_catalogue & self.have_guide_star_catalogue:
            raise ValueError("We must have a targets, guide star and standard star catalogue! Did you forget to load them?")

        # Add the empty columns to our dataframes
        self._add_columns_to_catalogues()
        self._rename_columns() ## ToDo: FixMe so this isn't necessary

        starting_tile = 0 
        if not self.config['fresh_start']:
            starting_tile = self.find_premade_tiles()
        

        # Now do the tiling
        self.logger.info(f"Total Targets in field to tile: {len(self.df_targets[self.df_targets['ALREADY_TILED']==False])}")
        N_tiles = 9999
        for current_tile in range(starting_tile, N_tiles):

            if self.df_targets['ALREADY_TILED'].sum() == len(self.df_targets):
                self.logger.info("\n\n")
                self.logger.info("Done!")
                break

            tile_out_fname = os.path.expanduser(f"{self.config['output_folder']}/Tiles/tile_{current_tile:03}.fld")
            guide_tile_out_fname = os.path.expanduser(f"{self.config['output_folder']}/Tiles/guide_tile_{current_tile:03}.fld")
            tile_out_fname_after_DC = f"{self.config['output_folder']}/DistortionCorrected/DC_tile_{current_tile:03}.fld"
            guide_tile_out_fname_after_DC = f"{self.config['output_folder']}/DistortionCorrected/guide_DC_tile_{current_tile:03}.fld"

            # filename for a plot which compares the tile before/after distortion correction
            DC_correction_comparison_plot_filename = f"{self.distortion_corrected_plot_location}/Tile_{current_tile:03}_comparison.pdf"

            # If we're applying distortion correction, make sure that the configuration code is looking at the correct file
            if configure_tiles:
                if apply_distortion_correction:
                    tile_file_for_configuration = tile_out_fname_after_DC
                else:
                    tile_file_for_configuration = tile_out_fname

            self.logger.info(f"Tile {current_tile}:")

            # Run the tiling
            self.df_targets, tile_df, guide_stars_for_tile, standard_stars_for_tile, tile_RA, tile_Dec = tiling.make_best_tile(self.df_targets, self.df_guide_stars, self.df_standard_stars, proximity=self.config['proximity'], tiling_parameters=self.config, tiling_type=self.config['tiling_type'], selection_type=self.config['allocation_type'], fill_spares_with_repeats=self.config['fill_spares_with_repeats'])

            self.logger.info(f"\tSaving to {tile_out_fname}...")

            # And save the outputs
            tiling.save_tile_outputs(f"{self.config['output_folder']}", self.df_targets, tile_df, guide_stars_for_tile, standard_stars_for_tile, tile_RA, tile_Dec, tiling_parameters=self.config, tile_number=current_tile, plot=True)

            if apply_distortion_correction == True:
                # Run the distortion correction
                DC_corrected_output_file = self.apply_distortion_correction(tile_out_fname, guide_tile_out_fname, tile_out_fname_after_DC, guide_tile_out_fname_after_DC, tile_RA, tile_Dec, guide_stars_for_tile, plot_save_filename=DC_correction_comparison_plot_filename)
                self.logger.info(f"\tDistortion-corrected tile saved at {tile_out_fname_after_DC}...")


            if configure_tiles == True:
                # Place holder variable to see if the configuration code has completed
                configured = False
                for j in range(self.config['MAX_TRIES']):

                    # If we're having trouble with a tile, rescale the proximity by 10 % for every five tiles we've tried
                    if self.config['Rescale_proximity'] is True:
                        proximity = self.config['proximity'] * (1 + (j // 5) * 0.1)

                    # Now call Caro's code
                    self.logger.info(f"Tile {current_tile}: \n\tRunning Configuration code on file {tile_out_fname_after_DC}...")
                    Configuration_bash_code = ["Rscript", f"{self.ConfigurationCode_location}", f"{tile_file_for_configuration}", f'{self.config["output_filename_stem"]}_', '--out-dir', f'{self.config["output_folder"]}/Configuration/']
                    try:
                        process = subprocess.run(Configuration_bash_code, text=True, capture_output=True, timeout=config_timeout)
                        self.logger_R_code.info(process.stdout)
                        return_code = process.returncode
                    except subprocess.TimeoutExpired:
                        return_code = 10
                        pass
                        
                    # If we've done the tile, break out of the inner 'Max tries' loop
                    if return_code == 0:
                        configured = True
                        break
                    elif return_code == 10:
                        timeout_message = f"\t**Tiling code couldn't configure after 10 minutes... Trying again with a new input tile.**\n\t**Proximity value is currently {proximity}**"
                        self.logger.info(timeout_message)
                        self.logger_R_code.info(timeout_message)
                    else:
                        tiling_error_message = "\t***Error in the tiling code! Exiting to debug***\n"
                        self.logger_R_code.info(process.stderr)
                        self.logger_R_code.error(tiling_error_message)
                        self.logger.info(tiling_error_message)
                        raise subprocess.CalledProcessError(return_code, Configuration_bash_code)
                        

                    # If we get here, it means the configuration code hasn't managed to configure. So we'll give it another tile. 
                    self.df_targets, tile_df, guide_stars_for_tile, standard_stars_for_tile, tile_RA, tile_Dec = tiling.make_best_tile(self.df_targets, self.df_guide_stars, self.df_standard_stars, proximity=proximity, tiling_parameters=self.config, tiling_type=self.config['tiling_type'], selection_type=self.config['allocation_type'], fill_spares_with_repeats=self.config['fill_spares_with_repeats'])

                    tiling.save_tile_outputs(f"{self.config['output_folder']}", self.df_targets, tile_df, guide_stars_for_tile, standard_stars_for_tile, tile_RA, tile_Dec, tiling_parameters=self.config, tile_number=current_tile, plot=True)

                    if apply_distortion_correction == True:
                        DC_corrected_output_file = self.apply_distortion_correction(tile_out_fname, guide_tile_out_fname, tile_out_fname_after_DC, guide_tile_out_fname_after_DC, tile_RA, tile_Dec, guide_stars_for_tile, plot_save_filename=DC_correction_comparison_plot_filename)


                if not configured:
                    raise ValueError(f"Couldn't configure this tile after {self.config['MAX_TRIES']} attempts.")

            # Now find which targets have been selected in this tile
            selected_targets_mask = self.df_targets['CATAID'].isin(tile_df['CATAID'])

            # Find which targets are being tiled for the first time
            new_targets = (selected_targets_mask) & (self.df_targets['ALREADY_TILED'] == False)
            # Change the TILED flag for targets we've added to a tile
            self.df_targets.loc[new_targets, 'ALREADY_TILED'] = True
            # And include the tile number for each target
            self.df_targets.loc[new_targets, 'Tile_number'] = current_tile

            self.best_tile_RAs.append(tile_RA)
            self.best_tile_Decs.append(tile_Dec)

            self.logger.info(f"FINISHED TILE {current_tile}\n\n")

        self.tile_positions = np.array([self.best_tile_RAs, self.best_tile_Decs])

        if plot:
            fig, ax = tiling.plot_survey_completeness_and_tile_positions(self.tile_positions, self.df_targets, self.config, fig=None, ax=None, completion_fraction_to_calculate=0.95, verbose=True)
            fig.savefig(f"{self.config['output_folder']}/Plots/{self.config['output_filename_stem']}_field_plot.pdf", bbox_inches='tight')
        
        #Save the overall dataframe
        self.df_targets.to_csv(f"{self.config['output_folder']}/Tiles/overall_targets_dataframe.csv")

        self.N_tiles = current_tile


    def apply_distortion_correction(self, tile_out_fname, guide_tile_out_fname, tile_out_fname_after_DC, guide_tile_out_fname_after_DC, tile_RA, tile_Dec, guide_stars_for_tile, verbose=False, plot_save_filename=None):

        """
        Take a tile file from the tiling process and apply Keith's distortion correction code. Then turn his one output file into the two output files which Caro's configuration code expects. We also need to do some work to edit the headers/etc.

        If plot_save_filename is not None, save an image of the tile before/after the correction is applied
        """

        # Now call Keith's code
        DC_bash_code = [f"{self.DistortionCorrection_binary_location}",  tile_out_fname, guide_tile_out_fname, tile_out_fname_after_DC]

        process = subprocess.Popen(DC_bash_code, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        output, error = process.communicate()
        if verbose:
            self.logger.info(output.decode("utf-8"))
        if process.returncode != 0:
            raise ValueError(output.decode("utf-8"))


        # Now take Keith's code and separate the one output file which contains galaxies, standard stars and guides into two: one which has galaxies and standard stars and one which just has the guides. This is a bit messy, but reflects the way that Caro's code reads things in. It's easier to do it in Python, where I know exactly which rows are guides/galaxies/standards, rather than guessing in R using the naming convention 

        with open(tile_out_fname_after_DC,"r") as fi:
            header = []
            for ln in fi:
                if ln.startswith("#"):
                    header.append(ln)

        header.insert(0, f"# Tile file after distortion correction applied\n")
        header.insert(1, f"# {tile_RA} {tile_Dec}\n")

        DC_corrected_output_file = pd.read_csv(tile_out_fname_after_DC, skiprows=6)
        guide_rows = DC_corrected_output_file.ID.isin(guide_stars_for_tile.CoADD_ID)
        with open(guide_tile_out_fname_after_DC, 'w') as f:
            for ln in header:
                f.write(ln)
            DC_corrected_output_file.loc[guide_rows].to_csv(f, index=False)

        with open(tile_out_fname_after_DC, 'w') as g:
            for ln in header:
                g.write(ln)
            DC_corrected_output_file.loc[~guide_rows].to_csv(g, index=False)

        if plot_save_filename is not None:
            fig, ax = plotting_tools.plot_distortion_correction_before_after(tile_out_fname_after_DC, title_text=f"{tile_out_fname_after_DC}")
            fig.savefig(f"{plot_save_filename}", bbox_inches='tight')
            plt.close(fig)
        
        return DC_corrected_output_file




    def allocate_hexabundles_for_single_tile(self, tile_number, plot=False):

        ### FIXME- add documentation here
        
        # fileNameGuides = ('GAMA_'+batch+'/Configuration/HECTORConfig_Guides_GAMA_'+batch+'_tile_%03d.txt' % (tileNum))
        fileNameGuides = f"{self.configuration_location}/HECTORConfig_Guides_{self.config['output_filename_stem']}_tile_{tile_number:03d}.txt"

        # proxy file to arrange guides in required format to merge with hexa probes
        proxyGuideFile = f'{self.allocation_files_location_base}/newfile.txt'

        # Adding ID column and getting rid of the header line of Guides cluster to add to the hexa cluster
        file_arranging.arrange_guidesFile(fileNameGuides, proxyGuideFile)

        # fileNameHexa = ('GAMA_'+batch+'/Configuration/HECTORConfig_Hexa_GAMA_'+batch+'_tile_%03d.txt' % (tileNum))
        fileNameHexa = f"{self.configuration_location}/HECTORConfig_Hexa_{self.config['output_filename_stem']}_tile_{tile_number:03d}.txt"

        plate_file = f"{self.allocation_files_location_base}/Hexa_and_Guides_{self.config['output_filename_stem']}_tile_{tile_number:03d}.txt"
        # plate_file = get_file('GAMA_'+batch+'/Output/Hexa_and_Guides_GAMA_'+batch+'_tile_%03d.txt' % (tileNum))

        #### Offset function: thermal coefficient based movement of magnet pair as a whole
        plate_file,magnetPair_offset = offsets.magnetPair_radialPositionOffset(plate_file)

        # Adding guides cluster txt file to hexa cluster txt file
        file_arranging.merge_hexaAndGuides(fileNameHexa, proxyGuideFile, plate_file)

        # extracting all the magnets and making a list of them from the plate_file
        all_magnets = extract_data.create_list_of_all_magnets_from_file(extract_data.get_file(plate_file), magnetPair_offset)

        #### Offset functions- still a work in progress- need to determine input source and add column to output file
        all_magnets = offsets.hexaPositionOffset(all_magnets)

        # create magnet pickup areas for all the magnets
        plots.create_magnet_pickup_areas(all_magnets)

        if plot:
            #************** # creating plots and drawing pickup areas
            plt.clf()
            plt.close()
            HECTOR_plate().draw_circle('r')
            plots.draw_magnet_pickup_areas(all_magnets, '--c')
            #**************


        # test for collision and detect magnets which have all pickup directions blocked
        conflicted_magnets = conflicts.functions.find_all_blocked_magnets(all_magnets)

        # create a list of the fully blocked magnets
        fully_blocked_magnets = conflicts.functions.create_list_of_fully_blocked_magnets(conflicted_magnets)

        fully_blocked_magnets_dictionary = conflicts.functions.blocking_magnets_for_fully_blocked_magnets(conflicted_magnets)

        # print the fully blocked magnets out in terminal and record in conflicts file
        conflictsRecord = f'{self.allocation_files_location_base}/Conflicts_Index.txt'
        conflicts.blocked_magnet.print_fully_blocked_magnets(fully_blocked_magnets,conflictsRecord, fileNameHexa)

        conflictFile = f'{self.allocation_files_location_base}/unresolvable_conflicts.txt'
        flagsFile = f'{self.allocation_files_location_base}/Flags.txt'
        #***  Choose former method OR median method OR larger bundle prioritized method for hexabundle allocation  ***
        positioning_array,self.galaxyIDrecord = position_ordering.create_position_ordering_array(all_magnets, fully_blocked_magnets, conflicted_magnets, self.galaxyIDrecord, self.config['output_filename_stem'], tile_number, conflictFile, flagsFile)

        if plot:
            # draw all the magnets in the plots created earlier
            figureFile = f"{self.plot_location}/savedPlot_{self.config['output_filename_stem']}_tile_{tile_number:03d}.pdf"
            plots.draw_all_magnets(all_magnets, self.config['output_filename_stem'], tile_number,figureFile)  #***********

        # checking positioning_array prints out all desired parameters
        print(positioning_array)

        # insert column heading and print only rectangular magnet rows in the csv file
        newrow = ['Magnet', 'Label', 'Center_x', 'Center_y', 'rot_holdingPosition', 'rot_platePlacing', 'order', 'Pickup_option', 'ID','Index', 'Hexabundle']
        newrow_circular = ['Magnet', 'Label', 'Center_x', 'Center_y', 'holding_position_ang', 'plate_placement_ang', 'order', 'Pickup_option', 'ID', 'Index', 'Hexabundle']

        # final two output files
        outputFile = f"{self.allocation_files_location_tiles}/Hexa_and_Guides_with_PositioningArray_{self.config['output_filename_stem']}_tile_{tile_number:03d}.txt"
        robotFile = f"{self.allocation_files_location_tiles}/Robot_{self.config['output_filename_stem']}_tile_{tile_number:03d}.txt"

        # creating robotFile array and storing it in robot file
        positioning_array, robotFilearray = file_arranging.create_robotFileArray(positioning_array,robotFile,newrow,fully_blocked_magnets_dictionary)

        # adjusting the positioning array to merge only selected parameters to the output file
        positioning_array, positioning_array_circular = file_arranging.positioningArray_adjust_and_mergetoFile(positioning_array, plate_file, outputFile, newrow,newrow_circular)

        # produce final files with consistent layout and no extra commas
        file_arranging.finalFiles(outputFile, robotFile)

        # just to check each tile's whole operation time
        # print("\t \t -----   %s seconds   -----" % (time.time() - start_time))

        # Comment out all ***** marked plot functions above(lines 81-84,105s)
        # to run whole batch of tiles fast without plots


    def allocate_hexabundles_for_all_tiles(self):

        for tile_number in range(self.N_tiles):
            self.allocate_hexabundles_for_single_tile(tile_number)



import matplotlib.pyplot as plt
import matplotlib.patches as patches
from ...general_operations.trigonometry import rotational_matrix,convert_degrees_to_radians
from ...hector.constants import robot_center_x,robot_center_y
from math import pi, cos, sin

class rectangle:

    # initializing a rectangle shape with the respective parameters
    def __init__(self,center,length,width,orientation):
        self.center = center
        self.length = length
        self.width = width
        self.orientation = orientation

        # fixing plotting view for Robot file coordinates
        self.view_y = self.center[0]
        self.view_x = - self.center[1]
        # self.view_orientation = self.orientation# - 90

    # calculating four corners of the rectangular shape
    def calculate_4corners(self):

        rotation_matrix = rotational_matrix(convert_degrees_to_radians(self.orientation))

        self.corner1 = (self.center[0] - rotation_matrix[0][0] * self.width / 2 - rotation_matrix[1][0] * self.length / 2, \
                        self.center[1] - rotation_matrix[0][1] * self.width / 2 - rotation_matrix[1][1] * self.length / 2)

        self.corner2 = (self.center[0] + rotation_matrix[0][0] * self.width / 2 - rotation_matrix[1][0] * self.length / 2, \
                        self.center[1] + rotation_matrix[0][1] * self.width / 2 - rotation_matrix[1][1] * self.length / 2)

        self.corner3 = (self.center[0] + rotation_matrix[0][0] * self.width / 2 + rotation_matrix[1][0] * self.length / 2, \
                        self.center[1] + rotation_matrix[0][1] * self.width / 2 + rotation_matrix[1][1] * self.length / 2)

        self.corner4 = (self.center[0] - rotation_matrix[0][0] * self.width / 2 + rotation_matrix[1][0] * self.length / 2, \
                        self.center[1] - rotation_matrix[0][1] * self.width / 2 + rotation_matrix[1][1] * self.length / 2)

        return [self.corner1,self.corner2,self.corner3,self.corner4]

    # plotting the rectangle on respective figures
    def draw_rectangle(self,colour):

        rectangle.calculate_4corners(self)

        plt.figure(1)
        plt.plot([self.corner1[0], self.corner2[0]], [self.corner1[1], self.corner2[1]], colour,alpha = 0.5,zorder=3)
        plt.plot([self.corner2[0], self.corner3[0]], [self.corner2[1], self.corner3[1]], colour,alpha = 0.5,zorder=3)
        plt.plot([self.corner3[0], self.corner4[0]], [self.corner3[1], self.corner4[1]], colour,alpha = 0.5,zorder=3)
        plt.plot([self.corner4[0], self.corner1[0]], [self.corner4[1], self.corner1[1]], colour,alpha = 0.5,zorder=3)
        plt.axis('scaled')

################################################################
        plt.figure(2)
        plt.plot([self.corner1[0] + robot_center_x, self.corner2[0] + robot_center_x], [self.corner1[1] + robot_center_y, self.corner2[1] + robot_center_y], colour)
        plt.plot([self.corner2[0] + robot_center_x, self.corner3[0] + robot_center_x], [self.corner2[1] + robot_center_y, self.corner3[1] + robot_center_y], colour)
        plt.plot([self.corner3[0] + robot_center_x, self.corner4[0] + robot_center_x], [self.corner3[1] + robot_center_y, self.corner4[1] + robot_center_y], colour)
        plt.plot([self.corner4[0] + robot_center_x, self.corner1[0] + robot_center_x], [self.corner4[1] + robot_center_y, self.corner1[1] + robot_center_y], colour)
        plt.axis('scaled')
################################################################

    # plotting the four corners of rectangle on respective figures
    def plot_corners(self):

        rectangle.calculate_4corners(self)

        plt.figure(1)
        plt.plot(self.corner1[0], self.corner1[1], 'xb')
        plt.plot(self.corner2[0], self.corner2[1], 'xb')
        plt.plot(self.corner3[0], self.corner3[1], 'xb')
        plt.plot(self.corner4[0], self.corner4[1], 'xb')

#############################################
        plt.figure(2)
        plt.plot(self.corner1[0] + robot_center_x, self.corner1[1] + robot_center_y, 'xb')
        plt.plot(self.corner2[0] + robot_center_x, self.corner2[1] + robot_center_y, 'xb')
        plt.plot(self.corner3[0] + robot_center_x, self.corner3[1] + robot_center_y, 'xb')
        plt.plot(self.corner4[0] + robot_center_x, self.corner4[1] + robot_center_y, 'xb')
###############################################

    # plotting the corner labels on respective figures
    def label_corners_on_plot(self):

        rectangle.calculate_4corners(self)

        plt.figure(1)
        plt.text(self.corner1[0], self.corner1[1], str(1), fontsize=15)
        plt.text(self.corner2[0], self.corner2[1], str(2), fontsize=15)
        plt.text(self.corner3[0], self.corner3[1], str(3), fontsize=15)
        plt.text(self.corner4[0], self.corner4[1], str(4), fontsize=15)

################################################################
        plt.figure(2)
        plt.text(self.corner1[0], self.corner1[1], str(1), fontsize=15)
        plt.text(self.corner2[0], self.corner2[1], str(2), fontsize=15)
        plt.text(self.corner3[0], self.corner3[1], str(3), fontsize=15)
        plt.text(self.corner4[0], self.corner4[1], str(4), fontsize=15)
################################################################



import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import csv
import re
import time

# arranging the guides files in consistent format with hexas files
def arrange_guidesFile(fileNameGuides,proxyGuideFile):

    # index to keep count
    ExtraCols = 1

    # reading the guides cdv file
    df = pd.read_csv(fileNameGuides)

    # The columns being filled up with NaN to keep the guides file same format as the hexas file
    for i in range(35):
        df['ColFillUp%d' % ExtraCols] = 'NaN'
        ExtraCols += 1

    # writing to the proxy guides file
    df.to_csv(proxyGuideFile, index=False, sep=' ', quoting=csv.QUOTE_NONE, escapechar=' ')

    # adding in two more columns at the start of the proxy file for the IDs and the magnet indexes
    rows = [line.split(' ') for line in open(proxyGuideFile)]
    cols = list(zip(*rows))
    cols.insert(1, ['"1stCol"', 'NaN', 'NaN', 'NaN', 'NaN', 'NaN', 'NaN'])
    cols.insert(1, ['21', '22', '23', '24', '25', '26', '27'])
    cols.insert(1, ['"IDs"', 'NaN', 'NaN', 'NaN', 'NaN', 'NaN', 'NaN'])
    cols = cols[1:]
    rows = list(zip(*cols))

    # writing those two columns created above to the file
    with open(proxyGuideFile, 'w') as f:
        f.writelines([' '.join(row) for row in rows[1:]])

# merging the hexas and guides file to create one plate file with all the magnets
def merge_hexaAndGuides(fileNameHexa, proxyGuideFile, plate_file):

    # adding the proxy guides file to the hexas file
    with open(fileNameHexa) as fp:
        file1 = fp.read()
    with open(proxyGuideFile) as fp:
        file2 = fp.read()
    file1 += file2

    # creating a plate file to write the added hexas and guides
    with open(plate_file, 'w+') as fp:
        fp.write(file1)

    # print statement to check from terminal the plate file of the tile which is being checked
    print("Filename being checked for conflicts:")
    print(fileNameHexa)

# creating the robotFile array for
def create_robotFileArray(positioning_array,robotFile,newrow,fully_blocked_magnets_dictionary):

    # guide probes do not have ID, so they are allocated a large negative integer
    positioning_array[:, 8] = [i if i != 'nan' else -999999 for i in positioning_array[:, 8]]
    for i in range(len(positioning_array[:, 8])):
        positioning_array[:, 8][i] = round(float(positioning_array[:, 8][i]))

    # an array for the robot file is created which is sorted based on the placement order numbering
    robotFilearray = sorted(positioning_array, key=lambda x: x[6])

    # the title row is inserted at the first row of the array
    robotFilearray = np.insert(robotFilearray, 0, np.array(newrow), 0)

    # add the reposition column to robot file by using the fully blocked magnets dictionary
    robotFilearray = add_repositionCol_to_robotFile(positioning_array,robotFilearray,fully_blocked_magnets_dictionary)

    # TEST PRINT
    print(robotFilearray)

    # write the robot file array into the CSV file for the robot
    with open(robotFile, 'w+') as robotFile:
        writer = csv.writer(robotFile, delimiter=' ')
        writer.writerows(robotFilearray)

    return positioning_array,robotFilearray

def add_repositionCol_to_robotFile(positioning_array,robotFilearray,fully_blocked_magnets_dictionary):

    # Creates a list containing w lists, each of h item/s, all filled with 0
    w, h = len(positioning_array[:, 8]) + 1, 1
    nameColumn = [['[0]' for x in range(w)] for y in range(h)]

    j=0
    for i in robotFilearray:
        if hasNumbers(i[1]):
            nameColumn[0][j] = i[1]
        else:
            nameColumn[0][j] = i[1] + str(i[9])
        j += 1

    print(nameColumn)

    # transposing the list to a column with a title assigned
    nameColumn[0][0] = 'Magnet_title'
    nameColumn = np.transpose(nameColumn)

    # add the 'list' column to the robot file array in desired position
    robotFilearray = np.hstack((robotFilearray[:, :8], nameColumn, robotFilearray[:, 8:]))

    # Creates a list containing w lists, each of h item/s, all filled with 0
    w, h = len(positioning_array[:, 9]) + 1, 1
    rePosition_col = [['[0]' for x in range(w)] for y in range(h)]

    # TEST PRINT to check the arrays
    print(len(positioning_array[:, 8]))
    print(len(rePosition_col))
    print(len(nameColumn))

    # filling out the created list with the blocked magnets dictionary in order with the robot file array
    for each_magnet in fully_blocked_magnets_dictionary:
        for i in range(len(robotFilearray)):
            if (robotFilearray[i][0] + ' ' + str(robotFilearray[i][10])) == each_magnet:
                for j in range(len(fully_blocked_magnets_dictionary[each_magnet])):
                    if j>0:
                        rePosition_col[0][i] += '_'
                    else:
                        rePosition_col[0][i] = ''
                    for k in range(len(robotFilearray)):
                        if robotFilearray[k][0]+' '+str(robotFilearray[k][10]) == fully_blocked_magnets_dictionary[each_magnet][j]:
                            rePosition_col[0][i] += str(robotFilearray[k][8])

    # transposing the list to a column with a title assigned
    rePosition_col[0][0] = 'rePosition_magnets'
    rePosition_col = np.transpose(rePosition_col)

    ## TEST PRINTs
    print(rePosition_col)
    print(rePosition_col.shape)
    print(robotFilearray.shape)

    # add the 'list' column to the robot file array in desired position
    robotFilearray = np.hstack((robotFilearray[:, :9], rePosition_col, robotFilearray[:, 9:]))

    return robotFilearray

def hasNumbers(inputString):
    return any(char.isdigit() for char in inputString)

def positioningArray_adjust_and_mergetoFile(positioning_array, plate_file, outputFile, newrow,newrow_circular):

    # splitting positioning array into circular
    positioning_array_circular = np.vsplit(positioning_array, 2)[0]

    # adding the title row and getting rid of some columns not required
    positioning_array_circular = np.insert(positioning_array_circular, 0, np.array(newrow_circular), 0)
    positioning_array_circular = np.delete(positioning_array_circular, [2, 3, 4, 5, 9, 10], 1)

    # splitting positioning array to keep only rectangular ones
    positioning_array = np.vsplit(positioning_array, 2)[1]

    # adding the title row and getting rid of some columns not required
    positioning_array = np.insert(positioning_array, 0, np.array(newrow), 0)
    positioning_array = np.delete(positioning_array, [2, 3, 4, 5, 8], 1)

    # index for keeping count
    index = 0

    # Open the input_file in read mode and output_file in write mode
    with open(plate_file, 'r') as read_obj, \
            open(outputFile, 'w+', newline='') as write_obj:

        # Create a csv.reader object from the input file object
        csv_reader = csv.reader(read_obj)

        # Create a csv.writer object from the output file object
        csv_writer = csv.writer(write_obj)

        # Read each row of the input csv file as list
        for row in csv_reader:
            roww_circular = np.array2string(positioning_array_circular[index], separator=' ',formatter={'str_kind': lambda x: x})
            roww = np.array2string(positioning_array[index], separator=' ', formatter={'str_kind': lambda x: x})

            # Append the default text in the row / list
            row.append(str(roww).replace('[', ' ').rstrip(']'))
            row.append(str(roww_circular).replace('[', ' ').rstrip(']'))

            # Add the updated row / list to the output file
            csv_writer.writerow(row)

            # updating index count
            index += 1

    return positioning_array,positioning_array_circular

# final files being formatted to maintain consistency
def finalFiles(outputFile, robotFile):

    df3 = pd.read_csv(outputFile, header=0)
    df3.to_csv(outputFile, index=False, sep=' ', quoting=csv.QUOTE_NONE, escapechar=' ')

    # df4 = pd.read_csv(robotFile, header=0, error_bad_lines=False)
    # df4.to_csv(robotFile, index=False, sep=' ', quoting=csv.QUOTE_NONE, escapechar=' ')



import yaml
from pathlib import Path

# Load the file with all the constant values in. We need to do some work to get the paths correct...
# This file is 3 folders deep from the 'hop' folder, so we need to go three "parents" up
location_of_module = Path(__file__).parent.parent.parent
with open(f"{location_of_module}/constants/HECTOR_CONSTANTS.yaml", 'r') as f:
    constants_dict = yaml.safe_load(f)

circular_magnet_radius                     = constants_dict['circular_magnet_radius']
rectangle_magnet_width                     = constants_dict['rectangle_magnet_width']
rectangle_magnet_length                    = constants_dict['rectangle_magnet_length']
circular_rectangle_magnet_distance         = constants_dict['circular_rectangle_magnet_distance']
circular_rectangle_magnet_center_distance  = (circular_magnet_radius + circular_rectangle_magnet_distance
                                              + rectangle_magnet_length / 2.0)
HECTOR_plate_center_coordinate             = constants_dict['HECTOR_plate_center_coordinate']
HECTOR_plate_radius                        = constants_dict['HECTOR_plate_radius'] # previous value=260.0
robot_arm_length                           = constants_dict['robot_arm_length']
robot_arm_width                            = constants_dict['robot_arm_width']
circular_magnet_pickuparea_length          = robot_arm_length
circular_magnet_pickuparea_width           = (3.0 * robot_arm_width / 2.0) + circular_magnet_radius
rectangle_magnet_pickuparea_length         = robot_arm_length

if (robot_arm_width < ((rectangle_magnet_length - robot_arm_width) / 2)):
    rectangle_magnet_pickuparea_width = (robot_arm_width + rectangle_magnet_length) / 2
elif (robot_arm_width >= ((rectangle_magnet_length - robot_arm_width) / 2)):
    rectangle_magnet_pickuparea_width = 2 * robot_arm_width
#rectangle_magnet_pickuparea_width          = 0.5 * (rectangle_magnet_length + 3.0 * robot_arm_width)

# robot file center coordinates, depending on the mechanical mounting of the plate
robot_center_x = constants_dict['robot_center_x']
robot_center_y = constants_dict['robot_center_y']


from ..general_operations.trigonometry import rotational_matrix,convert_degrees_to_radians, convert_radians_to_degrees
from ..hector.constants import robot_center_x,robot_center_y
from math import atan,sin,cos
import numpy as np
import pandas as pd
import string
import sys
import re

# Adjusting offset to move circular magnet closer to OR far from rectangular magnet
def hexaPositionOffset(all_magnets):

    offset_distance = 0.0  # to be derived from excel file

    for i in all_magnets:

        if i.__class__.__name__ == 'rectangular_magnet' and i.index == 8:

            # getting the orientation of rectangular magnet with respect to North(up)
            ang = i.orientation

            # check to adjust angle within 0 to -360 range
            if ang < 360:
                ang = ang + 360

            # print('orientation_rectangular='+str(ang))
            azAngs = convert_radians_to_degrees(i.azAngs)
            # print('azAngs_rectangular'+str(azAngs))
            ang_azAngs = convert_radians_to_degrees(i.rectangular_magnet_input_orientation)
            # print('ang_azAngs_rectangular' + str(ang_azAngs))


    for i in all_magnets:

        if i.__class__.__name__ == 'circular_magnet' and i.index == 8:

            # adjusting the angle to ensure movement is about the rectangular magnet's centre axis
            angle_adjusted = 450+ang

            # check to adjust angle within 0 to 360 range
            if angle_adjusted > 360:
                angle_adjusted = angle_adjusted - 360

            rotation_matrix_circle = rotational_matrix(convert_degrees_to_radians(angle_adjusted))
            # subtracting offset distance moves circular magnet closer to rectangular
            i.center = (i.center[0] - rotation_matrix_circle[0][0] * offset_distance, \
                                      i.center[1] - rotation_matrix_circle[0][1] * offset_distance)
            i.offset = offset_distance

            # print('orientation_circular=' + str(ang))
            azAngs = convert_radians_to_degrees(i.azAngs)
            # print('azAngs_circular=' + str(azAngs))


    return all_magnets


# thermal expansion related offset which will move the magnet pair as a whole based on certain coefficients
def magnetPair_radialPositionOffset(plate_file):

    ## read off a certain table from csv file or derived off some equations and calculations to identify the magnet pair
    ## to be adjusted with the determined offset values, VARIABLE 'offset_distance' used for testing function's usability

    offset_radialDistance = 20  # to be derived

    # T_observed > T_configured then radial inward movement by magnet
    # T_observed < T_configured then radial outward movement by magnet

    # store magnet pair index and offset distance accordingly, to be derived
    magnetPair_offset = []
    # magnetPair_offset = [(14,-30),(4,-30),(12,-30),(9,-30)] # +ve value makes radial outward movement, and -ve value for radial inward movement

    return plate_file, magnetPair_offset

# radial Position offset being adjusted in the extract_data.py file before all_magnets are being produced
def radialPositionOffset(list_of_probes,magnetPair_offset):

    # iterating through each magnet in the offset list and the probes list
    for item in magnetPair_offset:
        for each_probe in list_of_probes:

            # looking for match of items in offset and probes list
            if item[0] == each_probe.index:

                ### TEST PRINT
                print(each_probe.circular_magnet_center)

                # calculating the angle of the magnet with respect to x-axis
                theta = atan(each_probe.circular_magnet_center[1] / each_probe.circular_magnet_center[0])

                # calculating sine and cosine angle adjustment to offset distance for the radial movement
                # x values in positive range and -ve range move in opposite radial directions, so this step ensure same direction movement
                if each_probe.circular_magnet_center[0] >= 0:
                    each_probe.circular_magnet_center = (each_probe.circular_magnet_center[0] + (cos(theta) * item[1]), \
                                                         each_probe.circular_magnet_center[1] + (sin(theta) * item[1]))
                else:
                    each_probe.circular_magnet_center = (each_probe.circular_magnet_center[0] - (cos(theta) * item[1]), \
                                                         each_probe.circular_magnet_center[1] - (sin(theta) * item[1]))

                ### TEST PRINT
                print(each_probe.circular_magnet_center)

    return list_of_probes


from setuptools import setup, find_packages, Command
from setuptools.command.install import install
import distutils.command.install as orig
from distutils.command.build import build
import subprocess
import inspect
import pathlib
from glob import glob

# Man this took a while. I want to run make to build the Distortion Correction code automatically. So I've made a custom build class which runs "build" and then runs make
# However, this creates all the files we need but they're in a temporary directory! They're not copied into the install directory 
# I'm therefore calling them "data" and telling setuptools to move them
# BUT: by default, data isn't moved to the same location as the install files, for some reason. We therefore use a hack to get the location of the installation files (which is probably something like lib/python3.7/site-packages/)
# This is a horrible hack... And 100% shouldn't be necessary. There's a "package_data" option in setuptools for just this kind of circumstance. But I couldn't get it to work- it doesn't seem to do what it's supposed to. This is a common question online and doesn't seem to be well documented __at all__
import pip , os
installation_location = os.path.split(os.path.split(pip.__file__)[0])[0]
print(f"installation_location: {installation_location}")

class CustomBuild(build):

  def run(self):
        # run original build code
        build.run(self)


        def compile():
          subprocess.call(['make', 'all_clean', '-C', 'hop/distortion_correction/HectorTranslationSoftware/Code'])
          subprocess.call(['make', '-C', 'hop/distortion_correction/HectorTranslationSoftware/Code'])

        self.execute(compile, [], 'Compiling Distortion Correction binary')




# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()


data_files = [(f'/{installation_location}/hop/distortion_correction/HectorTranslationSoftware/Code', ['hop/distortion_correction/HectorTranslationSoftware/Code/HectorConfigUtil'] + glob('hop/distortion_correction/HectorTranslationSoftware/Code/*.o')), 
      (f'/{installation_location}/hop/distortion_correction/HectorTranslationSoftware/Packages/Misc', glob('hop/distortion_correction/HectorTranslationSoftware/Packages/Misc/*.o')), 
      (f'/{installation_location}/hop/distortion_correction/HectorTranslationSoftware/Packages/sds', glob('hop/distortion_correction/HectorTranslationSoftware/Packages/sds/*.o')),
      (f'/{installation_location}/hop/distortion_correction/HectorTranslationSoftware/Packages/DramaErs', glob('hop/distortion_correction/HectorTranslationSoftware/Packages/DramaErs/*.o')),
      (f'/{installation_location}/hop/distortion_correction/HectorTranslationSoftware/Packages/slalib_o', glob('hop/distortion_correction/HectorTranslationSoftware/Packages/slalib_o/*.o'))
      ]



setup(name='Hector-Observations-Pipeline',
      version='0.3.2',
      description='Hector Galaxy Survey Observations pipeline',
      long_description=README,
      long_description_content_type="text/markdown",
      author='Sam Vaughan',
      author_email='sam.vaughan@sydney.edu.au',
      url='https://github.com/samvaughan/Hector-Observations-Pipeline',
      license="GNU",
      classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8"],
      packages=find_packages(),
      include_package_data = True,
      data_files=data_files,
      python_requires='>=3',
      cmdclass={'build': CustomBuild}
     )




from ...general_operations.geometry_shapes.circle import circle
from ...hector.constants import circular_magnet_radius,robot_arm_width
from math import sin, cos, pi
from ...hector.magnets.pickup_areas import tangential_right,tangential_left,radial_inward,radial_outward

class circular_magnet(circle):

    # initializing a circular magnet with all respective parameters
    def __init__(self,center,orientation,index,galaxyORstar,Re,mu_1re,Mstar,magnet_label,hexabundle,rads,rotation_pickup,rotation_putdown,azAngs,rectangular_magnet_input_orientation,IDs):

        radius = circular_magnet_radius
        super().__init__(center,radius,orientation)
        self.index = index
        self.placement_index = 0
        self.galaxyORstar = galaxyORstar
        self.Re = Re
        self.mu_1re = mu_1re
        self.Mstar = Mstar
        self.magnet_label = magnet_label
        self.hexabundle = hexabundle
        self.rads = rads
        self.rotation_pickup = rotation_pickup
        self.rotation_putdown = rotation_putdown
        self.azAngs = azAngs
        self.IDs = IDs

    # calculating the distance between magnet center to pickup area center
    def calculate_center_magnet_to_center_pickup_area_length(self):

        center_magnet_to_center_pickup_area_length = 0.5 * (0.5 * robot_arm_width + self.radius)

        return center_magnet_to_center_pickup_area_length

    # calculating the center coordinate of pickup area- Tangential Right
    def calculate_center_coordinate_tangential_right_pickuparea(self):

        center_coordinates = \
        [self.center[0] + self.calculate_center_magnet_to_center_pickup_area_length() * sin((pi / 180) * (- self.orientation)),
         self.center[1] + self.calculate_center_magnet_to_center_pickup_area_length() * cos((pi / 180) * (- self.orientation))]

        return center_coordinates

    # calculating the center coordinate of pickup area- Tangential Left
    def calculate_center_coordinate_tangential_left_pickuparea(self):

        center_coordinates = \
        [self.center[0] - self.calculate_center_magnet_to_center_pickup_area_length() * sin((pi / 180) * (- self.orientation)),
         self.center[1] - self.calculate_center_magnet_to_center_pickup_area_length() * cos((pi / 180) * (- self.orientation))]

        return center_coordinates

    # calculating the center coordinate of pickup area- Radial Inward
    def calculate_center_coordinates_radial_inward_pickuparea(self):

        center_coordinates = \
        [self.center[0] + self.calculate_center_magnet_to_center_pickup_area_length() * sin((pi / 180) * (90 - self.orientation)),
         self.center[1] + self.calculate_center_magnet_to_center_pickup_area_length() * cos((pi / 180) * (90 - self.orientation))]

        return center_coordinates

    # calculating the center coordinate of pickup area- Radial Outward
    def calculate_center_coordinates_radial_outwards_pickuparea(self):

        center_coordinates = \
        [self.center[0] - self.calculate_center_magnet_to_center_pickup_area_length() * sin((pi / 180) * (90 - self.orientation)),
         self.center[1] - self.calculate_center_magnet_to_center_pickup_area_length() * cos((pi / 180) * (90 - self.orientation))]

        return center_coordinates

    # calculating pickup areas using their respective center coordinates
    def create_pickup_areas(self):

        self.pickup_areas = \
            [tangential_right(self.calculate_center_coordinate_tangential_right_pickuparea(), -self.orientation),
             tangential_left( self.calculate_center_coordinate_tangential_left_pickuparea(),  -self.orientation),
             radial_inward(   self.calculate_center_coordinates_radial_inward_pickuparea(),   270 - self.orientation), # (270 -) added for adjustment of the 10mm RAW
             radial_outward(  self.calculate_center_coordinates_radial_outwards_pickuparea(), 270 - self.orientation)] # (270 -) added for adjustment of the 10mm RAW

        return self.pickup_areas

# check for the magnet being circular type
def is_circular_magnet(magnet):
    return isinstance(magnet, circular_magnet)

